# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import codecs
from os.path import join, exists, isfile, isdir
from OVSConstants import OVS_REPO_VERSION, OVS_REPO_HEAD, OVS_REPO_META
from OVSServerPool import get_owning_manager_uuid
from OVSDiscoverCommon import SimpleDiscoveryXML, DiscoveryLock

def validate_path(path):
    if os.path.exists(path):
        return True
    return False


def get_repo_meta_data(repo_dir):
    d = {}
    repometa_file = join(repo_dir, OVS_REPO_META)
    if not exists(repometa_file):
        return d

    f = codecs.open(repometa_file, encoding='utf-8')
    try:
        for line in f:
            line = line.strip()
            if not line:
                continue
            if line.startswith("#"):
                continue
            k, v = line.split("=", 1)
            d[k] = v
    finally:
        f.close()

    return d



def validate_repo_manager(repo_dir):

    d = get_repo_meta_data(repo_dir)

    if d:
        manager_uuid = d["OVS_REPO_MGR_UUID"]

        if manager_uuid == get_owning_manager_uuid():
            return True

    return False


class RepositoryInfo:

    def __init__(self, mount_point, doc):

        self.name = mount_point
        self.path = join(OVS_REPO_HEAD, mount_point)
        self.doc = doc


    def general_info(self):

        d = get_repo_meta_data(self.path)

        self.doc.push("Repository", "Name", self.name)

        if d:
            if d.has_key("OVS_REPO_VERSION"):
                self.doc.add("Version", d["OVS_REPO_VERSION"])
            if d.has_key("OVS_REPO_MGR_UUID"):
                self.doc.add("Manager_UUID",d["OVS_REPO_MGR_UUID"])
            if d.has_key("OVS_REPO_UUID"):
                self.doc.add("Repository_UUID",d["OVS_REPO_UUID"])
            if d.has_key("OVS_REPO_ALIAS"):
                self.doc.add("Repository_Alias", d["OVS_REPO_ALIAS"])




    def discover_and_generate_xml(self, mount_path, topdir_1, subdir_1, \
                                        topdir_2, subdir_2, item_name):


        path = join(mount_path, topdir_1)
        if not os.path.exists(path):
            return

        names = os.listdir(path)
        self.doc.push(topdir_1)
        if not names:
            self.doc.pop()
            return None

        if subdir_1:
            for name in names:
                if name.startswith(".") or isfile(name):
                    continue
                self.doc.push(subdir_1, "Name", name)
                subpath = join(path, name)
                if subdir_2:
                    self.discover_and_generate_xml(subpath, topdir_2, subdir_2, \
                                                    None, None, item_name)
                    self.doc.pop()
                    continue

                files = os.listdir(subpath)
                for filename in files:
                    if filename.startswith(".") or isdir(filename):
                        continue
                    self.doc.add(item_name, filename)

                self.doc.pop()

            self.doc.pop()
            return

        for name in names:
            if name.startswith(".") or isdir(name):
                continue
            self.doc.add(item_name, name)

        self.doc.pop()
        return

def discover_repositories(args):

    """
    Discover information about the repositories and associated info.

    @param      List of mounted file systems

    @return     Returns the discovery data as an XML document.
                Raises an exception on failure.
    """

    lock = DiscoveryLock("discover_repositories")
    if not lock.acquire():
        raise Exception("discover_repositories already in progress")

    if not validate_path(OVS_REPO_HEAD):
        lock.release()
        raise Exception("Unable to discover repository: %s" % OVS_REPO_HEAD)

    mountpoints = args.split()

    repolist = []

    repo_mountpoints = os.listdir(OVS_REPO_HEAD)

    selected_mountpoints = list(set(mountpoints).intersection(set(repo_mountpoints)))

    doc = SimpleDiscoveryXML("Discover_Repositories_Result")

    doc.push("RepositoryList")

    for entry in selected_mountpoints:
        if os.path.ismount(join(OVS_REPO_HEAD, entry)):
            repolist.append(RepositoryInfo(entry, doc))

    for repo in repolist:

        repo.general_info()

        repo.discover_and_generate_xml(repo.path, "Assemblies", "Assembly", \
                                                 None, None, "File")
        repo.discover_and_generate_xml(repo.path, "Templates", "Template", \
                                                 None, None, "File")
        repo.discover_and_generate_xml(repo.path, "ISOs", None, \
                                                 None, None, "ISO")
        repo.discover_and_generate_xml(repo.path, "VirtualDisks", None, \
                                                 None, None, "Disk")
        #repo.discover_and_generate_xml(repo.path, "ServerPools", "ServerPool", \
        #                                         "VirtualMachines", "VirtualMachine", "File")

        repo.discover_and_generate_xml(repo.path, "VirtualMachines", "VirtualMachine", \
                                                 None, None, "File")

        doc.pop()

    doc.pop()
    try:
        result = doc.finish()
    except Exception, e:
        lock.release()
        raise Exception("Unable to format discover_repositories XML: %s" % e)

    lock.release()
    return result
