# Copyright (C) 2006-2011 Oracle. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

import os
import codecs
import commands
from os.path import join, exists, islink
from OVSConstants import LOCAL_REPO_DB
from OVSCommons import get_uuid
from OVSAgentDB import write_item, dump_db, delete_item
from OVSConstants import OVS_REPO_VERSION, OVS_REPO_HEAD, OVS_REPO_META, OVS_REPO_BASE
from OVSServerPool import get_owning_manager_uuid, get_membership_state
from OVSDiscoverRepositories import get_repo_meta_data, validate_repo_manager
from OVSDiscoverCommon import SimpleDiscoveryXML, DiscoveryLock
from OVSDownload import ovs_download_file


class Repository(object):
    """ Base class for Storage Repos. """

    def __init__(self, repo_uuid, fs_location, repo_alias, repo_version, manager_uuid,
                 repo_mount_point, filesystem, name_label='Untitled', name_description='Storage Repository'):
        self.repo_uuid = repo_uuid
        self.fs_location = fs_location
        self.repo_alias = repo_alias
        self.repo_version = repo_version
        self.manager_uuid = manager_uuid
        self.repo_mount_point = repo_mount_point
        self.filesystem = filesystem
        self.type = "Repository"
        self.name_label = name_label
        self.name_description = name_description

    def mount(self):
        if os.path.ismount(self.repo_mount_point):
            return
        cmd = "mount -t %s %s %s" % \
            (self.filesystem, self.fs_location, self.repo_mount_point)
        (status, output) = commands.getstatusoutput(cmd)
        if status != 0:
            raise Exception(output)

    def mount_and_check(self):
        if not os.path.ismount(self.repo_mount_point):
            if not exists(self.repo_mount_point):
                try:
                    os.mkdir(self.repo_mount_point)
                except:
                    return "Unmounted"

            cmd = "mount -t %s %s %s" % \
                (self.filesystem, self.fs_location, self.repo_mount_point)
            (status, output) = commands.getstatusoutput(cmd)
            if status != 0:
                os.rmdir(self.repo_mount_point)
                return "Unmounted"

        dirs = os.listdir(self.repo_mount_point)
        repo_items = ["Assemblies", "Templates", "ISOs", \
             "VirtualDisks", "VirtualMachines", ".ovsrepo"]
        empty = True
        for item in repo_items:
            if item in dirs:
                empty = False

        if empty:
            return "Empty"

        d = get_repo_meta_data(self.repo_mount_point)
        if not d or not d.has_key("OVS_REPO_VERSION") or d["OVS_REPO_VERSION"] != '3.0':
            return "Invalid"

        return "Mounted"

    def unmount(self):
        if not os.path.ismount(self.repo_mount_point):
            return
        cmd = "umount %s" % self.repo_mount_point
        (status, output) = commands.getstatusoutput(cmd)
        if status != 0:
            raise Exception("%s:%s" % (cmd,output))

    def create(self):
        if not os.path.ismount(self.repo_mount_point):
            raise Exception("Repository not mounted")

        subdirs = ["Assemblies", "Templates", "ISOs", \
             "VirtualDisks", "VirtualMachines"]
        try:
            for entry in subdirs:
                path = join(self.repo_mount_point, entry)
                if not exists(path):
                    if islink(path):
                        os.remove(path)
                    os.mkdir(path)
            self.init_metadata()
        except:
            self.delete(True)
            raise

    def delete(self, erase=False):
        if erase:
            if not os.path.ismount(self.repo_mount_point):
                raise Exception("Repository not mounted")
            cmd = "rm -rf %s/%s;rm -rf %s/*" % (self.repo_mount_point, OVS_REPO_META, self.repo_mount_point )
            (status, output) = commands.getstatusoutput(cmd)
            if status != 0:
                raise Exception(output)

    def add(self):
        data = get_repo_meta_data(self.repo_mount_point)
        if not data:
            raise Exception("Repository is not valid")

        if data["OVS_REPO_MGR_UUID"] != "NONE" and self.manager_uuid != data["OVS_REPO_MGR_UUID"]:
            raise Exception("Repository not owned by server's manager")

        if OVS_REPO_VERSION != data["OVS_REPO_VERSION"]:
            raise Exception("Repository not correct version")

        self.repo_alias = data["OVS_REPO_ALIAS"]
        self.repo_version = OVS_REPO_VERSION
        self.repo_uuid = str(data["OVS_REPO_UUID"])

    def init_metadata(self):
        repometa_file = join(self.repo_mount_point, OVS_REPO_META)
        file = codecs.open(repometa_file, "w", encoding='utf-8')
        try:
            file.write("OVS_REPO_UUID=%s\nOVS_REPO_MGR_UUID=%s\nOVS_REPO_VERSION=%s\nOVS_REPO_ALIAS=%s" \
                       % (self.repo_uuid, self.manager_uuid, self.repo_version, self.repo_alias))
        finally:
            file.close()


def get_fs_type(fs_location):
    fs_location_type = get_fs_location_type(fs_location)
    if fs_location_type == "OVS_LUN":
        return "ocfs2"
    else:
        return "nfs"


def get_fs_location_type(fs_location):
    if ":" in fs_location:
        return "OVS_NFS"
    if fs_location.startswith("/dev/"):
        return "OVS_LUN"


def find_repo(repo_uuid):
    db = dump_db(LOCAL_REPO_DB)
    for repo_uuid_db, repo in db.items():
        if repo_uuid_db == repo_uuid:
            return repo
    return None


def mount_repository_fs(fs_location, mount_point):
    """
    Create the specified mount point directory under /OVS/Repositories
    and mount the repository file system on it.

    @param       fs_location - scsi disk or nfs share. Scsi disk will be
                              identified as a dm device.
                 mount_point - mount point of file system.
    """
    if not exists(OVS_REPO_BASE):
        os.mkdir(OVS_REPO_BASE)
        os.mkdir(OVS_REPO_HEAD)
    repo_mount_point = join(OVS_REPO_HEAD, mount_point)
    if os.path.ismount(repo_mount_point):
        return
    os.mkdir(repo_mount_point)
    filesystem = get_fs_type(fs_location)
    cmd = "mount -t %s %s %s" % \
        (filesystem, fs_location, repo_mount_point)
    (status, output) = commands.getstatusoutput(cmd)
    if status != 0:
        os.rmdir(repo_mount_point)
        raise Exception(output)


def unmount_repository_fs(mount_point):
    """
    Unmount the repository file system from its mount point and delete
    the mount point under /OVS/Repositories.

    @param       mount_point - mount point of file system.
    """
    repo_mount_point = join(OVS_REPO_HEAD, mount_point)
    if not os.path.ismount(repo_mount_point):
        raise Exception("Repository file system not mounted")
    cmd = "umount %s" % repo_mount_point
    (status, output) = commands.getstatusoutput(cmd)
    if status != 0:
        raise Exception(output)
    os.rmdir(repo_mount_point)


def create_repository(fs_location, mount_point, repo_uuid, repo_alias):
    """
    Create the repository and meta data file.

    @param      fs_location - scsi disk or nfs share. Scsi disk
                             will be identified as a dm device.
                mount_point - mount point of repository file system.
                             ( /OVS/Repositories/mount_point1/ )
                repo_uuid - Uuid of repository
                repo_alias - Friendly name of repository
    """
    repo = find_repo(repo_uuid)
    if repo:
        raise Exception("Repository already exists")
    repo_mount_point = join(OVS_REPO_HEAD, mount_point)
    repometa_file = join(repo_mount_point, OVS_REPO_META)
    if exists(repometa_file):
        raise Exception("File system contains repository data")
    filesystem = get_fs_type(fs_location)
    manager_uuid = get_owning_manager_uuid()
    repo = Repository(repo_uuid, fs_location, repo_alias, OVS_REPO_VERSION,
                      manager_uuid, repo_mount_point, filesystem)
    repo.create()
    write_item(LOCAL_REPO_DB, repo_uuid, repo)


def add_repository(fs_location, mount_point):
    """
    Add an existing repository to the server.

    @param      fs_location - scsi disk or nfs share. Scsi disk
                             will be identified as a dm device.
                mount_point - mount point of repository file system.
                             ( /OVS/Repositories/mount_point1/ )
    """
    repo_mount_point = join(OVS_REPO_HEAD, mount_point)

    if not os.path.ismount(repo_mount_point):
        raise Exception("Repository not mounted")

    filesystem = get_fs_type(fs_location)
    manager_uuid = get_owning_manager_uuid()
    repo = Repository(None, fs_location, None, None, \
                             manager_uuid, repo_mount_point, filesystem)
    repo.add()
    added_repo = find_repo(repo.repo_uuid)
    if added_repo:
        return
    write_item(LOCAL_REPO_DB, repo.repo_uuid, repo)


def delete_repository(repo_uuid, erase=False):
    """
    Delete the repository and meta data file.

    @param      repo_uuid - Uuid of repos used as mount point
                erase - If true delete the repos subdirs, otherwise just
                          unmount from server and delete the mount point
    """
    repo = find_repo(repo_uuid)
    if not repo:
        raise Exception("Repository not found in repository.db")
    if repo.manager_uuid != "NONE" and repo.manager_uuid != get_owning_manager_uuid():
        raise Exception("Repository not owned by server manager")
    repo.delete(erase)
    delete_item(LOCAL_REPO_DB, repo_uuid)


def get_repository_meta_data(path):
    """
    Return contents of meta data file.

    @param      path - full path of meta data directory.
    """
    return get_repo_meta_data(path)


def synch_repositories( ):
    """
    Read repo db and mount any unmounted repos or unmount any missing repos.
    """
    mounted_repolist = []
    db_repolist = []

    file = open("/proc/mounts")
    for line in file:
        fs_spec, fs_file, fs_vfstype = line.split()[0:3]
        if fs_file.startswith(OVS_REPO_HEAD):
            mounted_repolist.append(fs_file)

    db = dump_db(LOCAL_REPO_DB)

    for repo_uuid, repo in db.items():
        db_repolist.append(repo.repo_mount_point)

    umount_set = list(set(mounted_repolist).difference(set(db_repolist)))
    mount_set = list(set(db_repolist).difference(set(mounted_repolist)))

    for item in umount_set:
        cmd = "umount %s" % item
        (status, output) = commands.getstatusoutput(cmd)
        if status != 0:
            raise Exception("Could not unmount repository %s" % item)

        if exists(item):
            os.rmdir(item)

    for repo_uuid, repo in db.items():
        if repo.repo_mount_point in mount_set:
            if not exists(repo.repo_mount_point):
                os.mkdir(repo.repo_mount_point)
            repo.mount()


def get_repo_free_space(repo_uuid):
    """
    Get free space of repo.

    @param      repo_uuid - Uuid of repository
    @return     "success:free_space=N" where N is the number of bytes"
    """
    repo = find_repo(repo_uuid)

    if not repo:
        raise Exception("Repository does not exist")

    if repo.manager_uuid != get_owning_manager_uuid():
        raise Exception("Repository not owned by server manager")

    if not os.path.ismount(repo.repo_mount_point):
        raise Exception("Repository not mounted")

    stat = os.statvfs(repo.repo_mount_point)

    return "success:free_space=%s" % (stat.f_bsize * stat.f_bfree)


def discover_repository_db():
    """
    Discover information about the contents of the repository data base.
    @return     Returns the discovery data as an XML document.
                Raises an exception on failure.
    """
    lock = DiscoveryLock("discover_repository_db")
    if not lock.acquire():
        raise Exception("discover_repository db already in progress")

    doc = SimpleDiscoveryXML("Discover_Repository_Db_Result")
    doc.push("RepositoryDbList")
    db = dump_db(LOCAL_REPO_DB)
    for repo_uuid_db, repo in db.items():
        doc.push("Repository", "Uuid", repo_uuid_db)
        doc.add("Version", repo.repo_version)
        doc.add("Fs_location", repo.fs_location)
        doc.add("Alias", repo.repo_alias)
        doc.add("Manager_uuid", repo.manager_uuid)
        doc.add("Mount_point", repo.repo_mount_point)
        doc.add("Filesystem_type", repo.filesystem)

        status = repo.mount_and_check()        # checks mount and repo directory
        if status == "Empty":
            delete_item(LOCAL_REPO_DB, repo.repo_uuid)
            repo.unmount()
            os.rmdir(repo.repo_mount_point)

        doc.add("Status", status)
        doc.pop()

    doc.pop()
    try:
        result = doc.finish()
    except Exception, e:
        lock.release()
        raise Exception("Unable to format discover_repository_db XML: %s" % e)

    lock.release()
    return result


def edit_repository_db(repo_uuid, db_changes):
    """
    edit the contents of the repository data base.

    @param      repo_uuid - Uuid of repository.
                db_entries - dictionary of values to modify.
    @return     Raises an exception on failure.
    """
    repo = find_repo(repo_uuid)
    if not repo:
        raise Exception("Repository not found in repository.db")
    if repo.manager_uuid != "NONE" and repo.manager_uuid != get_owning_manager_uuid():
        raise Exception("Repository not owned by server manager UUID %s" % get_owning_manager_uuid())
    delete_item(LOCAL_REPO_DB, repo_uuid)
    if db_changes.get("Version", None):
        repo.repo_version = db_changes["Version"]
    if db_changes.get("Alias", None):
        repo.repo_alias = db_changes["Alias"]
    if db_changes.get("Manager_uuid", None):
        repo.manager_uuid = db_changes["Manager_uuid"]
    repo.init_metadata()
    write_item(LOCAL_REPO_DB, repo_uuid, repo)


def validate_repo(repo):
    if not repo:
        raise Exception("Repository does not exist")
    if not os.path.ismount(repo.repo_mount_point):
        raise Exception("Repository not mounted")
    if not validate_repo_manager(repo.repo_mount_point):
        raise Exception("Repository is not owned by manager")


def import_file(repo_id, file_type, file_id, url, option, obj):
    repo = find_repo(repo_id)
    validate_repo(repo)
    filename = join(repo.repo_mount_point, file_type, file_id)
    ovs_download_file(url, filename, option, obj)


def delete_file(repo_id, file_type, file_id):
    repo = find_repo(repo_id)
    validate_repo(repo)
    filename = join(repo.repo_mount_point, file_type, file_id)
    if exists(filename):
        os.unlink(filename)


def import_virtual_disk(url, virtual_disk_id, repo_id, option=None):
    import_file(repo_id, 'VirtualDisks', virtual_disk_id, url, option, 'virtual_disk')


def import_ISO(url, iso_id, repo_id, option=None):
    import_file(repo_id, 'ISOs', iso_id, url, option, 'iso')


def delete_virtual_disk(virtual_disk_id, repo_id):
    delete_file(repo_id, 'VirtualDisks', virtual_disk_id)


def delete_ISO(iso_id, repo_id):
    delete_file(repo_id, 'ISOs', iso_id)


def create_virtual_disk(virtual_disk_id, repo_id, size_kbytes, sparse=True):
    repo = find_repo(repo_id)
    validate_repo(repo)
    filename = join(repo.repo_mount_point, "VirtualDisks", virtual_disk_id)
    if exists(filename):
        raise Exception("Virtual Disk %s already exists" % filename)
    (block_count, remainder) = divmod(size_kbytes, 1024)
    if remainder:
        block_count += 1
    if sparse:
        cmd = "dd if=/dev/zero of=%s bs=1M oflag=direct count=0 seek=%s" % (filename, block_count)
    else:
        cmd = "dd if=/dev/zero of=%s bs=1M oflag=direct count=%s" % (filename, block_count)
    (status, output) = commands.getstatusoutput(cmd)
    if status != 0:
        raise Exception("Error creating virtual disk %s: %s" % (filename, output))
